from dsge.DSGE import DSGE

import pkg_resources

# nkmp
relative_loc = 'examples/nkmp/'
model_file = pkg_resources.resource_filename('dsge', relative_loc+'nkmp.yaml')
data_file = pkg_resources.resource_filename('dsge', relative_loc+'us.txt')

nkmp = DSGE.read(model_file)
nkmp['__data__']['estimation']['data']['file'] = data_file

# sw
relative_loc = 'examples/sw/'
model_file = pkg_resources.resource_filename('dsge', relative_loc+'sw.yaml')
data_file = pkg_resources.resource_filename('dsge', relative_loc+'YY.txt')

sw = DSGE.read(model_file)
sw['__data__']['estimation']['data']['file'] = data_file



import sympy
import copy
from sympy.printing.str import StrPrinter

from sympy.core.cache import clear_cache

clear_cache()


StrPrinter._print_TSymbol = lambda self, x: x.__str__()


class Parameter(sympy.Symbol):
    def __init__(self, name, exp_date=0):
        super(Parameter, self).__init__()
        self.name = name

    # def __new__(self, name, exp_date=0):
    #     super(Parameter,self).__new__( name)

    def __repr__(self):
        return self.name

    def __set_prior(self, prior):
        self.prior = prior


class TSymbol(sympy.Symbol):

    # __xnew_cached_ = staticmethod(sympy.Symbol.__new_stage2__)

    def __init__(self, name, **args):
        super(TSymbol, self).__init__()

        if "date" not in args:
            self._assumptions["date"] = 0
            self.assumptions0["date"] = 0
        else:
            self._assumptions["date"] = args["date"]
            self.assumptions0["date"] = args["date"]
        if "exp_date" not in args:
            self._assumptions["exp_date"] = 0
            self.assumptions0["exp_date"] = 0
        else:
            self._assumptions["exp_date"] = args["exp_date"]
            self.assumptions0["exp_date"] = args["exp_date"]

        self._mhash = None
        self.__hash__()

        return None

    def __call__(self, lead):
        newdate = int(self.date) + int(lead)
        newname = str(self.name)
        # print 'creating', newname, newdate
        clear_cache()
        return self.__class__(newname, date=newdate)

    @property
    def date(self):
        return self.assumptions0["date"]

    @property
    def exp_date(self):
        return self.assumptions0["exp_date"]

    def _hashable_content(self):
        return (self.name, str(self.date), str(self.exp_date))

    def __getstate__(self):
        return {
            "date": self.date,
            "name": self.name,
            "exp_date": self.exp_date,
            "is_commutative": self.is_commutative,
            "_mhash": self._mhash,
        }

    def class_key(self):
        return (2, 0, self.name, self.date)

    @property
    def lag(self):
        return self.date

    def __str__(self):
        if self.lag == 0:
            result = self.name
        else:
            result = self.name + r"(" + str(self.lag) + r")"
        return result

    # def __repr__(self):
    #     return self.__str__()

    # def _repr_(self):
    #     return self.__str__()

    # def repr(self):
    #     return self.__str__()

    # def _print_TSymbol(self):
    #     return self.__str__()

    # def _print(self):
    #     return self.__str__()


class Variable(TSymbol):
    @property
    def fortind(self):
        if self.date <= 0:
            return "v_" + self.name
        else:
            return "v_E" + self.name

    def __str__(self):
        if self.exp_date == 0:
            result = super(Variable, self).__str__()
        else:
            result = "E[" + str(self.exp_date) + "]" + super(Variable, self).__str__()

        return result

    def __repr__(self):
        return self.__str__()

    __sstr__ = __str__


class LaggedExpectation(Variable):
    def __init__(self, name, date=0, exp_date=0):
        Variable.__init__(self, name, date)
        self.exp_date = exp_date

    def __getstate_(self):
        return {
            "date": self.date,
            "name": self.name,
            "is_commutative": self.is_commutative,
            "_mhash": self._mhash,
        }

    def _hashable_content(self):
        return (self.name, self.date, self.lag)


class Shock(TSymbol):
    @property
    def fortind(self):
        if self.date <= 0:
            return "e_" + self.name
        else:
            return "e_E" + self.name


class Equation(sympy.Equality):

    def __new__(cls, lhs, rhs, name=None):
        return super(sympy.Equality, cls).__new__(cls, lhs, rhs)

    @property
    def set_eq_zero(self):
        return self.lhs - self.rhs

    @property
    def variables(self):
        return [v for v in self.atoms() if isinstance(v, Variable)]



from __future__ import division

import numpy as np
from numpy.testing import assert_equal

from unittest import TestCase

from dsge import DSGE

from dsge.examples import sw

import pkg_resources

class TestOtherPara(TestCase):

    def test_simple(self):

        from dsge.examples import sw
        from dsge.translate import smc

        print(smc(sw))

        #self.assertAlmostEqual(1,0)



from __future__ import division

import numpy as np
from numpy.testing import assert_equal, assert_allclose

from unittest import TestCase

from dsge import DSGE

import pkg_resources

class TestFilter(TestCase):

    def test_sw_lik(self):
        from dsge.examples import sw

        sw = sw.compile_model()

        p0 = [0.1657,0.7869,0.5509,0.4312,0.1901,1.3333,1.6064,5.7606,0.72,0.7,1.9,0.65,0.57,0.3,0.5462,2.0443,0.8103,0.0882,0.2247,0.9577,0.2194,0.9767,0.7113,0.1479,0.8895,0.9688,0.5,0.72,0.85,0.4582,0.24,0.5291,0.4526,0.2449,0.141,0.2446]
        lik0 = sw.log_lik(p0)

        self.assertAlmostEqual(-829.7412615500879, lik0, places=6)

    def test_ar1(self):
        relative_loc = 'examples/ar1/'
        model_file = pkg_resources.resource_filename('dsge', relative_loc+'ar1.yaml')
        data_file = pkg_resources.resource_filename('dsge', relative_loc+'arma23_sim200.txt')
        ar1 = DSGE.DSGE.read(model_file)
        ar1['__data__']['estimation']['data'] = data_file

        ar1 = ar1.compile_model()

        rho = 0.85
        sigma = 1.00

        from scipy.stats import norm

        yy = np.asarray(ar1.yy).squeeze()
        yyhat = np.r_[0, rho*yy[:-1]]

        sig = sigma*np.ones_like(yyhat)
        sig[0] = sig[0]/np.sqrt(1.-rho**2)

        z = (yy - yyhat) / sig

        # Kalman Filter
        byhand = np.sum(norm.logpdf(yy, loc=yyhat, scale=sig))
        lik0 = ar1.log_lik([rho, sigma])
        self.assertAlmostEqual(byhand, lik0)


    def test_missing(self):
        relative_loc = 'examples/ar1/'
        model_file = pkg_resources.resource_filename('dsge', relative_loc+'ar1.yaml')
        data_file = pkg_resources.resource_filename('dsge', relative_loc+'arma23_sim200.txt')
        ar1 = DSGE.DSGE.read(model_file)
        ar1['__data__']['estimation']['data'] = data_file

        rho, sigma = ar1.p0()

        ar1 = ar1.compile_model()

        res = ar1.kf_everything([rho, sigma], y=np.nan*np.ones_like(ar1.yy))
        ll = res['log_lik'].sum().values[0]

        self.assertAlmostEqual(ll, 0)

        from scipy.stats import norm


        y = ar1.yy.values.copy().squeeze()

        sig = sigma*np.ones_like(y)
        sig[0] = sigma/np.sqrt(1.-rho**2)

        yyhat = np.r_[0, rho*y[:-1]]

        byhand = norm.logpdf(y[:-2], loc=yyhat[:-2], scale=sig[:-2]).sum()
        byhand += norm.logpdf(y[-1], loc=rho**2*y[-3], scale=np.sqrt(1+rho**2)*sig[-1])
        y[-2] = np.nan
        res = ar1.kf_everything([rho, sigma], y=y)
        ll = res['log_lik'].sum().values[0]

        self.assertAlmostEqual(ll, byhand)

    def test_filtered_values(self):
        from dsge.examples import nkmp as dsge

        p0 = dsge.p0()
        model = dsge.compile_model()

        res = model.kf_everything(p0)

        #print(res['filtered_states'])

        self.assertAlmostEqual(0, 0)


    def test_pred(self):
        relative_loc = 'examples/ar1/'
        model_file = pkg_resources.resource_filename('dsge', relative_loc+'ar1.yaml')
        data_file = pkg_resources.resource_filename('dsge', relative_loc+'arma23_sim200.txt')
        ar1 = DSGE.DSGE.read(model_file)
        ar1['__data__']['estimation']['data'] = data_file

        rho, sigma = ar1.p0()

        ar1 = ar1.compile_model()

        res = ar1.kf_everything([rho, sigma], y=ar1.yy)

        pred = ar1.pred([rho, sigma], shocks=False,h=5)

        y1 = ar1.yy.iloc[-1].values
        assert_allclose(pred, [rho*y1, rho**2*y1, rho**3*y1, rho**4*y1, rho**5*y1])



from scipy.stats import norm
from scipy.optimize import fsolve
import numpy as np
def d2Gammadomegadsigmafcn(z,sigma):
    f = (z/sigma-1)*norm.pdf(z);
    return f

def dGammadsigmafcn(z,sigma):
    f = -norm.pdf(z-sigma);
    return f


def zetaspbfcn(z,sigma,sprd):
    zetaratio = zetabomegafcn(z,sigma,sprd)/zetazomegafcn(z,sigma,sprd);
    nk = nkfcn(z,sigma,sprd);
    f = -zetaratio/(1-zetaratio)*nk/(1-nk);
    return f
        
def zetabomegafcn(z,sigma,sprd):
    nk = nkfcn(z,sigma,sprd);
    mustar = mufcn(z,sigma,sprd);
    omegastar = omegafcn(z,sigma);
    Gammastar = Gammafcn(z,sigma);
    Gstar = Gfcn(z,sigma);
    dGammadomegastar = dGammadomegafcn(z);
    dGdomegastar = dGdomegafcn(z,sigma);
    d2Gammadomega2star = d2Gammadomega2fcn(z,sigma);
    d2Gdomega2star = d2Gdomega2fcn(z,sigma);
    f = (omegastar*mustar*nk*(d2Gammadomega2star*dGdomegastar-d2Gdomega2star*dGammadomegastar)/
         (dGammadomegastar-mustar*dGdomegastar)**2/sprd/
         (1-Gammastar+dGammadomegastar*(Gammastar-mustar*Gstar)/(dGammadomegastar-mustar*dGdomegastar)))
    return f

def zetazomegafcn(z,sigma,sprd):
    mustar = mufcn(z,sigma,sprd);
    f = (omegafcn(z,sigma)*(dGammadomegafcn(z)-mustar*dGdomegafcn(z,sigma))/
    (Gammafcn(z,sigma)-mustar*Gfcn(z,sigma)));
    return f

def nkfcn(z,sigma,sprd):
    f = 1-(Gammafcn(z,sigma)-mufcn(z,sigma,sprd)*Gfcn(z,sigma))*sprd;
    return f

def mufcn(z,sigma,sprd):
    f = (1-1/sprd)/(dGdomegafcn(z,sigma)/dGammadomegafcn(z)*(1-Gammafcn(z,sigma))+Gfcn(z,sigma));
    return f

def omegafcn(z,sigma):
    f = np.exp(sigma*z-1/2*sigma**2);
    return f

def Gfcn(z,sigma):
    f = norm.cdf(z-sigma);
    return f

def Gammafcn(z,sigma):
    f = omegafcn(z,sigma)*(1-norm.cdf(z))+norm.cdf(z-sigma);
    return f

def dGdomegafcn(z,sigma):
    f = norm.pdf(z)/sigma;
    return f

def d2Gdomega2fcn(z,sigma):
    f = -z*norm.pdf(z)/omegafcn(z,sigma)/sigma**2;
    return f

def dGammadomegafcn(z):
    f = 1-norm.cdf(z);
    return f

def d2Gammadomega2fcn(z,sigma):
    f = -norm.pdf(z)/omegafcn(z,sigma)/sigma;
    return f

def dGdsigmafcn(z,sigma):
    f = -z*norm.pdf(z-sigma)/sigma;
    return f

def d2Gdomegadsigmafcn(z,sigma):
    f = -norm.pdf(z)*(1-z*(z-sigma))/sigma**2;
    return f

norminv = norm.ppf
from scipy.optimize import fsolve
def get_sigwstar(zwstar, sprd, zeta_spb):
    return fsolve(lambda x: zetaspbfcn(zwstar, x, sprd)-zeta_spb, 0.5)


import numpy as np
from numpy.testing import assert_equal

from unittest import TestCase

from dsge.DSGE import DSGE
from dsge.symbols import Variable, Equation

from numpy.testing import assert_equal

class TestEquations(TestCase):

    def test_equation(self):

        y = Variable('y')
        eq = Equation(y, 0.6*y(-1))
        print(y.date)
        var_list = [y]
        subs_dict = {}

        subs_dict.update({v:0 for v in var_list})
        subs_dict.update({v(-1):0 for v in var_list})
        subs_dict.update({v(1):0 for v in var_list})

        diff = eq.set_eq_zero.diff(y(-1)).subs(subs_dict)
        self.assertAlmostEqual(diff, -0.6)

        diff = eq.set_eq_zero.diff(y(1)).subs(subs_dict)
        self.assertAlmostEqual(diff, 0)

        diff = eq.set_eq_zero.diff(y).subs(subs_dict)
        self.assertAlmostEqual(diff, 1.0)



import numpy as np

from numba import jit


@jit(nopython=True)
def chand_recursion(y, CC, TT, RR, QQ, DD, ZZ, HH, A0, P0, t0=0):
    nobs, ny = y.shape

    At = A0
    Pt = P0

    loglh = 0.0

    Ft = ZZ @ Pt @ ZZ.T + HH
    Ft = 0.5 * (Ft + Ft.T)
    iFt = np.linalg.inv(Ft)

    St = TT @ Pt @ ZZ.T
    Mt = -iFt
    Kt = St @ iFt

    for i in range(nobs):
        yhat = ZZ @ At + DD.flatten()
        nut = y[i] - yhat

        dFt = np.log(np.linalg.det(Ft))
        iFtnut = np.linalg.solve(Ft, nut)

        if i >= t0:
            loglh = (
                loglh
                - 0.5 * ny * np.log(2 * np.pi)
                - 0.5 * dFt
                - 0.5 * np.dot(nut, iFtnut)
            )

        At = CC + TT @ At + Kt @ nut.T

        ZZSt = ZZ @ St
        MSpZp = Mt @ (ZZSt.T)
        TTSt = TT @ St

        Ft1 = Ft + ZZSt @ MSpZp
        # F_{t+1}
        Ft1 = 0.5 * (Ft1 + Ft1.T)
        iFt1 = np.linalg.inv(Ft1)

        Kt = (Kt @ Ft + TTSt @ MSpZp) @ iFt1
        # K_{t+1}
        St = TTSt - Kt @ ZZSt
        # S_{t+1}
        Mt = Mt + MSpZp @ iFt @ MSpZp.T
        # M_{t+1}
        Mt = 0.5 * (Mt + Mt.T)
        Ft = Ft1
        iFt = iFt1

    return loglh


@jit(nopython=True)
def kalman_filter(y, CC, TT, RR, QQ, DD, ZZ, HH, A0, P0, t0=0):

    # y = np.asarray(y)
    nobs, ny = y.shape
    ns = TT.shape[0]

    Pt = P0
    RQR = np.dot(np.dot(RR, QQ), RR.T)

    loglh = 0.0
    AA = np.zeros(shape=(ns))
    for i in range(nobs):

        not_missing = ~np.isnan(y[i])
        nact = not_missing.sum()
        yhat = ZZ[not_missing, :] @ AA + DD.flatten()[not_missing]

        nut = y[i][not_missing] - yhat

        Ft = (
            ZZ[not_missing, :] @ Pt @ ZZ[not_missing, :].T
            + HH[not_missing, :][:, not_missing]
        )
        Ft = 0.5 * (Ft + Ft.T)

        dFt = np.log(np.linalg.det(Ft))
        iFtnut = np.linalg.solve(Ft, nut)

        if i >= t0:
            loglh = (
                loglh
                - 0.5 * nact * np.log(2 * np.pi)
                - 0.5 * dFt
                - 0.5 * np.dot(nut, iFtnut)
            )

        TTPt = TT @ Pt
        Kt = TTPt @ ZZ[not_missing, :].T

        AA = CC + TT @ AA + Kt @ iFtnut
        Pt = TTPt @ TT.T - Kt @ np.linalg.solve(Ft, Kt.T) + RQR

    return loglh


@jit(nopython=True)
def filter_and_smooth(y, CC, TT, RR, QQ, DD, ZZ, HH, A0, P0, t0=0):

    nobs, ny = y.shape
    ns = TT.shape[0]

    At = A0
    Pt = P0
    RQR = np.dot(np.dot(RR, QQ), RR.T)

    forecast_means = np.zeros((nobs, ns))
    forecast_stds = np.zeros((nobs, ns))
    forecast_cov = np.zeros((nobs, ns, ns))

    filtered_means = np.zeros((nobs, ns))
    filtered_stds = np.zeros((nobs, ns))
    filtered_cov = np.zeros((nobs, ns, ns))

    smoothed_means = np.zeros((nobs, ns))
    smoothed_stds = np.zeros((nobs, ns))
    smoothed_cov = np.zeros((nobs, ns, ns))

    liks = np.zeros(nobs)

    Lmat = np.zeros((nobs, ns, ns))
    ZtiFtnut = np.zeros((nobs, ns))
    ZtiFtZ = np.zeros((nobs, ns, ns))

    for i in range(nobs):

        observed = ~np.isnan(y[i])
        nact = np.sum(observed)

        forecast_means[i] = At
        forecast_stds[i] = np.sqrt(np.diag(Pt))
        forecast_cov[i] = 0.5 * (Pt + Pt.T)

        if nact > 0:

            yhat = (ZZ @ At)[observed] + DD[observed].flatten()
            nut = y[i][observed] - yhat

            Ft = (ZZ @ Pt @ ZZ.T + HH)[observed, :][:, observed]
            Ft = 0.5 * (Ft + Ft.T)

            dFt = np.log(np.linalg.det(Ft))
            iFtnut = np.linalg.solve(Ft, nut)

            if i >= t0:
                liks[i] = (
                    - 0.5 * nact * np.log(2 * np.pi)
                    - 0.5 * dFt
                    - 0.5 * np.dot(nut, iFtnut)
                )

            Kt = Pt @ ZZ[observed, :].T

            Lmat[i] = TT- TT @ Pt @ ZZ[observed, :].T @ np.linalg.inv(Ft) @ ZZ[observed, :]
                       

            ZtiFtnut[i] = ZZ[observed, :].T @ iFtnut
            ZtiFtZ[i] = ZZ[observed, :].T @ np.linalg.inv(Ft) @ ZZ[observed, :]

            At1 = At + Kt @ iFtnut
            Pt1 = Pt - Kt @ np.linalg.solve(Ft, Kt.T)

        else:

            At1 = At
            Pt1 = Pt

        filtered_means[i] = At1
        filtered_stds[i] = np.sqrt(np.diag(Pt1))
        filtered_cov[i] = 0.5 * (Pt1 + Pt1.T)

        # forecast
        At = CC + TT @ At1
        Pt = TT @ Pt1 @ TT.T + RQR

    # smoother
    smoothed_means[-1] = filtered_means[-1]
    smoothed_stds[-1] = filtered_stds[-1]
    smoothed_cov[-1] = filtered_cov[-1]

    r = np.zeros((nobs + 1, ns))
    N = np.zeros((ns, ns))

    for i in range(nobs - 2, -1, -1):
        r[i - 1] = ZtiFtnut[i] + Lmat[i].T @ r[i]
        smoothed_means[i] = forecast_means[i] + forecast_cov[i] @ r[i - 1]

        N = ZtiFtZ[i] + Lmat[i].T @ N @ Lmat[i]
        smoothed_cov[i] = forecast_cov[i] - forecast_cov[i] @ N @ forecast_cov[i].T
        smoothed_stds[i] = np.sqrt(np.diag(smoothed_cov[i]))

    return (
        liks,
        filtered_means,
        filtered_stds,
        filtered_cov,
        forecast_means,
        forecast_stds,
        forecast_cov,
        smoothed_means,
        smoothed_stds,
        smoothed_cov,
    )



from __future__ import division

import numpy as np
from numpy.testing import assert_equal

from unittest import TestCase

from dsge import DSGE

import pkg_resources

class TestABCD(TestCase):

    def test_abcd(self):
        model = DSGE.DSGE.read('/home/eherbst/Dropbox/code/dsge/dsge/examples/pi/pi.yaml')

        m = model.compile_model()

        A,B,C,D = m.abcd_representation([1,1.2])

        v = np.linalg.eig(A - B @ np.linalg.inv(D) @ C)
        v = np.max(np.abs(v[0]))

        self.assertAlmostEqual(1.2, v, places=6)



import sympy
import yaml
from .symbols import Variable, Equation, Shock, Parameter, TSymbol

import scipy.stats as stats
from sympy.matrices import Matrix, zeros
import re
import numpy as np
import itertools
import pandas as p
from .StateSpaceModel import LinearDSGEModel

import warnings


class EquationList(list):
    """A container for holding an set of equations."""

    def __init__(self, args, context={}):
        self.context = context
        return super(EquationList, self).__init__(args)

    def __setitem__(self, key, value):

        if isinstance(value, str):
            lhs, rhs = value.split('=')

            lhs = eval(lhs, self.context)
            rhs = eval(rhs, self.context)

            value = Equation(lhs, rhs)

        return super(EquationList, self).__setitem__(key, value)


def read_data_file(datafile, obs_names):

    if type(datafile) == dict:
        startdate = datafile['start']
        datafile = datafile['file']
    else:
        startdate = 0

    try:
        with open(datafile, 'r') as df:
            data = df.read()
            delim_dict = {}

            if data.find(',') > 0:
                delim_dict['delimiter'] = ','

            data = np.genfromtxt(datafile, missing_values='NaN', **delim_dict)
    except:
        warnings.warn("%s could not be opened." % datafile)
        data = np.nan * np.ones((100, len(obs_names)))
        startdate = 0

    if len(obs_names) > 1:
        data = p.DataFrame(data[:, :len(obs_names)], columns=list(
            map(lambda x: str(x), obs_names)))
    else:
        data = p.DataFrame(data, columns=list(
            map(lambda x: str(x), obs_names)))

    if startdate is not 0:
        nobs = data.shape[0]
        data.index = p.period_range(startdate, freq='Q', periods=nobs)

    return data


def construct_prior(prior_list, parameters):

    prior = []
    for par in parameters:
        prior_spec = prior_list[par]

        ptype = prior_spec[0]
        pmean = prior_spec[1]
        pstdd = prior_spec[2]
        from scipy.stats import beta, norm, uniform, gamma
        from dsge.OtherPriors import InvGamma
        if ptype == 'beta':
            a = (1 - pmean) * pmean**2 / pstdd**2 - pmean
            b = a * (1 / pmean - 1)
            pr = beta(a, b)
            pr.name = 'beta'
            prior.append(pr)
        if ptype == 'gamma':
            b = pstdd**2 / pmean
            a = pmean / b
            pr = gamma(a, scale=b)
            pr.name = 'gamma'
            prior.append(pr)
        if ptype == 'normal':
            a = pmean
            b = pstdd
            pr = norm(loc=a, scale=b)
            pr.name = 'norm'
            prior.append(pr)
        if ptype == 'inv_gamma':
            a = pmean
            b = pstdd
            prior.append(InvGamma(a, b))
        if ptype == 'uniform':
            a, b = pmean, pstdd
            pr = uniform(loc=a, scale=(b - a))
            pr.name = 'uniform'
            prior.append(pr)

    return prior


class DSGE(dict):

    max_lead = 1;
    max_lag  = 1;

    def __init__(self, *kargs, **kwargs):
        super(DSGE, self).__init__(self, *kargs, **kwargs)

        fvars = []
        lvars = []

        # get forward looking variables
        for eq in self['equations']:

            variable_too_far = [v for v in eq.atoms() if isinstance(v, Variable) and v.date > 1]
            variable_too_early = [v for v in eq.atoms() if isinstance(v, Variable) and v.date < -1]

            eq_fvars = [v for v in eq.atoms() if isinstance(v, TSymbol) and v.date > 0]
            eq_lvars = [v for v in eq.atoms() if isinstance(v, TSymbol) and v.date < 0]

            fvars = list(set(fvars).union(eq_fvars))
            lvars = list(set(lvars).union(set(eq_lvars)))

        self['info']['nstate'] = len(self.variables) + len(fvars)

        self['fvars'] = fvars
        self['fvars_lagged'] = [Parameter('__LAGGED_'+f.name) for f in fvars]
        self['lvars'] = lvars
        self['re_errors'] = [Shock('eta_'+v.name) for v in self['fvars']]

        self['re_errors_eq'] = []
        i = 0
        for fv, lag_fv in zip(fvars, self['fvars_lagged']):
            self['re_errors_eq'].append(Equation(fv(-1) - lag_fv -  self['re_errors'][i], 0))
            i += 1

        if 'make_log' in self.keys():
            self['perturb_eq'] = []
            sub_dict = dict()
            sub_dict.update({v: Parameter(v.name + 'ss') * sympy.exp(v)
                            for v in self['make_log']})
            sub_dict.update({v(-1): Parameter(v.name + 'ss')
                            * sympy.exp(v(-1)) for v in self['make_log']})
            sub_dict.update({v(1): Parameter(v.name + 'ss') *
                            sympy.exp(v(1)) for v in self['make_log']})

            for eq in self.equations:
                peq = eq.subs(sub_dict)
                self['perturb_eq'].append(peq)

            self['ss_ordering'] = [Variable(v.name+'ss') for v in
                                   self['make_log']]

        else:
            self['perturb_eq'] = self['equations']

        # context = [(s.name, s) for s in self['par_ordering']+self['var_ordering']+self['shk_ordering']+self['para_func']]
        # context = dict(context)
        # context['log'] = sympy.log
        # context['exp'] = sympy.exp
        context = {}
        # self['pertub_eq'] = EquationList(self['perturb_eq'], context)

        return

    def __repr__(self):
        return "A DSGE Model."

    @property
    def equations(self):
        return self['equations']

    @property
    def variables(self):
        return self['var_ordering']

    @property
    def parameters(self):
        return [str(x) for x in self['par_ordering']]

    @property
    def shocks(self):
        return self['shk_ordering']

    @property
    def name(self):
        return self['name']

    @property
    def neq(self):
        return len(self['perturb_eq'])

    @property
    def neq_fort(self):
        return self.neq+self.neta

    @property
    def neta(self):
        return len(self['fvars'])

    @property
    def ns(self):
        return

    @property
    def ny(self):
        return len(self['observables'])

    @property
    def neps(self):
        return len(self['shk_ordering'])

    @property
    def npara(self):
        return len(self.parameters)

    def p0(self):
        return list(map(lambda x: self['calibration'][str(x)], self.parameters))

    def calibrate(**kwargs):
        pass
        
    def python_sims_matrices(self, matrix_format='numeric'):


        from sympy.utilities.lambdify import lambdify
        vlist = self['var_ordering'] + self['fvars']
        llist = [l(-1) for l in self['var_ordering']] + self['fvars_lagged']
        slist = self['shk_ordering']

        subs_dict = dict()

        eq_cond = self['perturb_eq'] + self['re_errors_eq']

        sub_var = self['var_ordering']
        subs_dict.update( {v:0 for v in sub_var})
        subs_dict.update( {v(1):0 for v in sub_var})
        subs_dict.update( {v(-1):0 for v in sub_var})


        svar = len(vlist)
        evar = len(slist)
        rvar = len(self['re_errors'])
        ovar = len(self['observables'])

        GAM0 = zeros(svar, svar)
        GAM1 = zeros(svar, svar)
        PSI  = zeros(svar, evar)
        PPI  = zeros(svar, rvar)

        eq_i = 0
        for eq in eq_cond:
            curr_var = filter(lambda x: x.date >= 0, eq.atoms(Variable))

            for v in curr_var:
                v_j = vlist.index(v)
                GAM0[eq_i, v_j] = -(eq).set_eq_zero.diff(v).subs(subs_dict)

            past_var = filter(lambda x: x in llist, eq.atoms())

            for v in past_var:
                v_j = llist.index(v)
                GAM1[eq_i, v_j] = eq.set_eq_zero.diff(v).subs(subs_dict)

            shocks = filter(lambda x: x, eq.atoms(Shock))

            for s in shocks:
                if s not in self['re_errors']:
                    s_j = slist.index(s)
                    PSI[eq_i, s_j] = eq.set_eq_zero.diff(s).subs(subs_dict)
                else:
                    s_j = self['re_errors'].index(s)
                    PPI[eq_i, s_j] = eq.set_eq_zero.diff(s).subs(subs_dict)

            eq_i += 1
            #print "\r Differentiating equation {0} of {1}.".format(eq_i, len(eq_cond)),
        DD = zeros(ovar, 1)
        ZZ = zeros(ovar, svar)

        eq_i = 0
        for obs in self['observables']:
            eq = self['obs_equations'][str(obs)]

            DD[eq_i, 0] = eq.subs(subs_dict)

            curr_var = filter(lambda x: x.date >= 0, eq.atoms(Variable))
            for v in curr_var:
                v_j = vlist.index(v)
                ZZ[eq_i, v_j] = eq.diff(v).subs(subs_dict)

            eq_i += 1


        if matrix_format=='symbolic':
            QQ = self['covariance']
            HH = self['measurement_errors']
            return GAM0, GAM1, PSI, PPI, QQ, DD, ZZ, HH
        #print ""
        from collections import OrderedDict
        subs_dict = []
        context = dict([(p, Parameter(p)) for p in self.parameters])
        context['exp'] = sympy.exp
        context['log'] = sympy.log
        context_f = {}
        context_f['exp'] = np.exp
        if 'helper_func' in self['__data__']['declarations']:
            from imp import load_source
            f = self['__data__']['declarations']['helper_func']['file']
            module = load_source('helper_func', f)
            for n in self['__data__']['declarations']['helper_func']['names']:
                context[n] = sympy.Function(n) #getattr(module, n)
                context_f[n] = getattr(module, n)
        #import sys
        #sys.stdout.flush()
        ss = {}

        for p in self['other_para']:
            ss[str(p)] = eval(str(self['para_func'][p.name]), context)
            context[str(p)] = ss[str(p)]
            #print("\r Constructing substitution dictionary [{0:20s}]".format(p.name),)
        #sys.stdout.flush()
        #print ""
        #context_f['numpy'] = 'numpy'
        GAM0 = lambdify([self.parameters+self['other_para']], GAM0)#, modules={'ImmutableDenseMatrix': np.array})#'numpy')
        GAM1 = lambdify([self.parameters+self['other_para']], GAM1)#, modules={'ImmutableDenseMatrix': np.array})#'numpy')
        PSI = lambdify([self.parameters+self['other_para']], PSI)#, modules={'ImmutableDenseMatrix': np.array})#'numpy')
        PPI = lambdify([self.parameters+self['other_para']], PPI)#, modules={'ImmutableDenseMatrix': np.array})#'numpy')

        psi = lambdify([self.parameters], [ss[str(px)] for px in self['other_para']])#, modules=context_f)

        def add_para_func(f):
            def wrapped_f(px):
                return f([*px, *psi(px)])
            return wrapped_f

        self.GAM0 = add_para_func(GAM0)
        self.GAM1 = add_para_func(GAM1)
        self.PSI = add_para_func(PSI)
        self.PPI = add_para_func(PPI)

        QQ = self['covariance'].subs(subs_dict)
        HH = self['measurement_errors'].subs(subs_dict)

        DD = DD.subs(subs_dict)
        ZZ = ZZ.subs(subs_dict)

        QQ = lambdify([self.parameters+self['other_para']], self['covariance'])
        HH = lambdify([self.parameters+self['other_para']], self['measurement_errors'])
        DD = lambdify([self.parameters+self['other_para']], DD)
        ZZ = lambdify([self.parameters+self['other_para']], ZZ)

        self.QQ = add_para_func(QQ)
        self.DD = add_para_func(DD)
        self.ZZ = add_para_func(ZZ)
        self.HH = add_para_func(HH)

        return GAM0, GAM1, PSI, PPI

    def compile_model(self):
        self.python_sims_matrices()

        GAM0 = self.GAM0
        GAM1 = self.GAM1
        PSI = self.PSI
        PPI = self.PPI

        QQ = self.QQ
        DD = self.DD
        ZZ = self.ZZ
        HH = self.HH

        if 'observables' not in self:
            self['observables'] = self['variables'].copy()
            self['obs_equations'] = dict(self['observables'],
                                         self['observables'])

        if 'data' in self['__data__']['estimation']:
            data = read_data_file(self['__data__']['estimation']['data'],
                                  self['observables'])
        else:
            data = np.nan * np.ones((100, len(self['observables'])))

        prior = None
        if 'prior' in self['__data__']['estimation']:
            prior = construct_prior(self['__data__']['estimation']['prior'],
                                    self.parameters)

        from .Prior import Prior as pri
        dsge = LinearDSGEModel(data, GAM0, GAM1, PSI, PPI,
                               QQ, DD, ZZ, HH, t0=0,
                               shock_names=list(map(str, self.shocks)),
                               state_names=list(map(str, self.variables+self['fvars'])),
                               obs_names=list(map(str, self['observables'])),
                               prior=pri(prior))

        return dsge

    def solve_model(self, p0):

        #if self.GAM0 == None:
        self.python_sims_matrices()

        from gensys import gensys_wrapper as gensys

        G0 = self.GAM0(*p0)
        G1 = self.GAM1(*p0)
        PSI = self.PSI(*p0)
        PPI = self.PPI(*p0)
        C0 = np.zeros((G0.shape[0]))
        TT, CC, RR, fmat, fwt, ywt, gev, RC, loose = gensys.call_gensys(G0, G1, C0, PSI, PPI, 1.00000000001)

        return TT, RR, RC

    def fix_parameters(self, **kwargs):

        for para, value in kwargs.items():
            para = Parameter(para)
            self.parameters.remove(para)
            self['other_para'].append(para)
            self['para_func'][str(para)] = value
        return self


    @classmethod
    def read(cls, mfile):
        """

        """

        f = open(mfile)
        txt = f.read()
        f.close()


        txt = txt.replace('^', '**')
        txt = txt.replace(';', '')
        txt = re.sub(r"@ ?\n", " ", txt)
        model_yaml = yaml.safe_load(txt)

        dec = model_yaml['declarations']
        cal = model_yaml['calibration']

        name = dec['name']

        var_ordering = [Variable(v) for v in dec['variables']]
        par_ordering = [Parameter(v) for v in dec['parameters']]
        shk_ordering = [Shock(v) for v in dec['shocks']]


        if 'para_func' in dec:
            other_para   = [Parameter(v) for v in dec['para_func']]
        else:
            other_para = []

        if 'observables' in dec:
            observables = [Variable(v) for v in dec['observables']]
            obs_equations = model_yaml['equations']['observables']
        else:
            observables = []
            obs_equations = dict()

        if 'measurement_errors' in dec:
            measurement_errors = [Shock(v) for v in dec['measurement_errors']]
        else:
            measurement_errors = None


        if 'make_log' in dec:
            make_log = [Variable(v) for v in dec['make_log']]
        else:
            make_log = []

        steady_state = [0]
        init_values = [0]

        context = [(s.name,s) for s in var_ordering + par_ordering + shk_ordering + other_para]
        context = dict(context)

        for f in [sympy.log, sympy.exp,
                  sympy.sin, sympy.cos, sympy.tan,
                  sympy.asin, sympy.acos, sympy.atan,
                  sympy.sinh, sympy.cosh, sympy.tanh,
                  sympy.pi, sympy.sign]:
            context[str(f)] = f

        context['sqrt'] = sympy.sqrt
        context['__builtins__'] = None

        equations = []

        if 'model' in model_yaml['equations']:
            raw_equations = model_yaml['equations']['model']
        else:
            raw_equations = model_yaml['equations']


        for eq in raw_equations:
            if '=' in eq:
                lhs, rhs = str.split(eq, '=')
            else:
                lhs, rhs = eq, '0'

            try:
                lhs = eval(lhs, context)
                rhs = eval(rhs, context)
            except TypeError as e:
                print('While parsing %s, got this error: %s' % (eq, repr(e)))
                return

            equations.append(Equation(lhs, rhs))



        #------------------------------------------------------------
        # Figure out max leads and lags
        #------------------------------------------------------------
        it = itertools.chain.from_iterable

        max_lead_exo = dict.fromkeys(shk_ordering)
        max_lag_exo = dict.fromkeys(shk_ordering)

        all_shocks = [list(eq.atoms(Shock)) for eq in equations]

        for s in shk_ordering:
            max_lead_exo[s] = max([i.date for i in it(all_shocks) if i.name == s.name])
            max_lag_exo[s] = min([i.date for i in it(all_shocks) if i.name == s.name])


        # arbitrary lags of exogenous shocks
        for s in shk_ordering:
            if abs(max_lag_exo[s]) > 0:
                var_s = Variable(s.name+"_VAR")
                var_ordering.append(var_s)
                equations.append(Equation(var_s, s))

                subs1 = [s(-i) for i in np.arange(1, abs(max_lag_exo[s])+1)]
                subs2 = [var_s(-i) for i in np.arange(1, abs(max_lag_exo[s])+1)]
                subs_dict = dict(zip(subs1, subs2))
                equations = [eq.subs(subs_dict) for eq in equations]


        all_vars = [list(eq.atoms(Variable)) for eq in equations]
        max_lead_endo = dict.fromkeys(var_ordering)
        max_lag_endo = dict.fromkeys(var_ordering)

        for v in var_ordering:
            max_lead_endo[v] = max([i.date for i in it(all_vars) if i.name == v.name])
            max_lag_endo[v] = min([i.date for i in it(all_vars) if i.name == v.name])

        #------------------------------------------------------------
        # arbitrary lags/leads of exogenous shocks
        subs_dict = dict()
        old_var = var_ordering[:]
        for v in old_var:

            # lags
            for i in np.arange(2, abs(max_lag_endo[v])+1):
                # for lag l need to add l-1 variable
                var_l = Variable(v.name + "_LAG" + str(i-1))

                if i == 2:
                    var_l_1 = Variable(v.name, date=-1)
                else:
                    var_l_1 = Variable(v.name + "_LAG" + str(i-2), date=-1)

                subs_dict[Variable(v.name, date=-i)] = var_l(-1)
                var_ordering.append(var_l)
                equations.append(Equation(var_l, var_l_1))


            # still need to do leads

        equations = [eq.subs(subs_dict) for eq in equations]

        cov = cal['covariances']

        nshock = len(shk_ordering)
        npara  = len(par_ordering)

        info = {'nshock': nshock, 'npara': npara}
        QQ = sympy.zeros(nshock, nshock)
        for key, value in cov.items():
            shocks = key.split(",")

            if len(shocks)==1:
                shocks.append(shocks[0])

            if len(shocks)==2:
                shocki = Shock(shocks[0].strip())
                shockj = Shock(shocks[1].strip())

                indi = shk_ordering.index(shocki)
                indj = shk_ordering.index(shockj)

                QQ[indi, indj] = eval(str(value), context)
                QQ[indj, indi] = QQ[indi, indj]

            else:
                "fdfadsf"

        nobs = len(obs_equations)
        HH = sympy.zeros(nobs, nobs)

        if measurement_errors is not None:
            for key, value in cal['measurement_errors'].items():
                shocks = key.split(",")

                if len(shocks) == 1:
                    shocks.append(shocks[0])

                if len(shocks) == 2:
                    shocki = Shock(shocks[0].strip())
                    shockj = Shock(shocks[1].strip())

                    indi = measurement_errors.index(shocki)
                    indj = measurement_errors.index(shockj)

                    HH[indi, indj] = eval(value, context)
                    HH[indj, indi] = HH[indi, indj]

        context['sum'] = np.sum
        context['range'] = range
        for obs in obs_equations.items():
            obs_equations[obs[0]] = eval(obs[1], context)

        calibration = model_yaml['calibration']['parameters']

        if 'parafunc' not in cal:
            cal['parafunc'] = {}

        model_dict = {
            'var_ordering': var_ordering,
            'par_ordering': par_ordering,
            'shk_ordering': shk_ordering,
            'other_parameters': other_para,
            'other_para': other_para,
            'para_func': cal['parafunc'],
            'calibration': calibration,
            'steady_state': steady_state,
            'init_values': init_values,
            'equations': equations,
            'covariance': QQ,
            'measurement_errors': HH,
            'meas_ordering': measurement_errors,
            'info': info,
            'make_log': make_log,
            '__data__': model_yaml,
            'name': dec['name'],
            'observables': observables,
            'obs_equations': obs_equations
            }

        model = cls(**model_dict)
        return model




import numpy as np
def EXP(j):
    return lambda x: E(j, x)

def E(j, x):
    if len(x.atoms()) > 1:

        for xa in x.atoms():
            if isinstance(xa, Variable):
                x = x.subs({xa:E(j, xa)})
        return x
    else:
        if isinstance(x, Parameter):
            return x
        if isinstance(x, Variable):
            return type(x)(x.name, date=x.date, exp_date=j)

def ASUM(x, d):
    return x

def cholpsd(x):
    n = x.shape[0]
    y = x.copy()
    for k in np.arange(0, n):
        if y[k, k]>0:
            y[k, k] = np.sqrt(y[k, k])
            if k < n-1:
                y[k+1:, k] = y[k+1:, k]/y[k, k]
                y[k, k+1:] = 0
                for j in np.arange(k+1, n):
                    y[j:, j] = y[j:, j] - y[j:, k]*y[j, k]

    return y



import numpy as np

import os

template_path = os.path.join(os.path.dirname(__file__), 'templates')

fortran_files = {'smc_driver': 'smc_driver_mpi.f90',
                 'rwmh_driver': 'rwmh_driver.f90', 
                 'blockmh_driver': 'blockmcmc.f90', 
                 'Makefile': 'Makefile_dsge'}

pdict = {'gamma': 1, 
         'beta': 2, 
         'norm': 3, 
         'inv_gamma': 4, 
         'uniform': 5}

fortran_model = """
module model_t
  use, intrinsic :: iso_fortran_env, only: wp => real64

  use gensys, only: do_gensys
  use fortress, only : fortress_lgss_model
  use fortress_prior_t, only: model_prior => prior

  implicit none

  type, public, extends(fortress_lgss_model) :: model
     integer :: neta

   contains
     procedure :: system_matrices
  end type model


  interface model
     module procedure new_model
  end interface model
  
  
contains

  type(model) function new_model() result(self)

    character(len=144) :: name, datafile, priorfile
    integer :: nobs, T, ns, npara, neps
    
    name = '{model[name]}'
    datafile = 'data.txt'
    priorfile = 'prior.txt'

    nobs = {yy.shape[1]}
    T = {yy.shape[0]}

    ns = {model.neq_fort}
    npara = {model.npara}
    neps = {model.neps}
    
    call self%construct_model(name, datafile, priorfile, npara, nobs, T, ns, neps)

!    self%p0 = {p0}
    self%neta = {model.neta}
    self%t0 = {t0}
  end function new_model

  subroutine system_matrices(self, para, error)

    class(model), intent(inout) :: self
    real(wp), intent(in) :: para(self%npara)

    integer, intent(out) :: error 
    
    ! double precision, intent(out) :: TT(self%ns, self%ns), RR(self%ns, neps), QQ(neps, neps)
    ! double precision, intent(out) :: DD(self%nobs), ZZ(self%nobs, self%ns), HH(self%nobs,self%nobs)
    double precision :: DD2(self%nobs,1)
    integer :: info

    double precision :: GAM0(self%ns, self%ns), GAM1(self%ns, self%ns), C(self%ns)
    double precision :: PSI(self%ns, self%neps), PPI(self%ns, self%neta), CC(self%ns)

    ! gensys
    double precision :: fmat, fwt, ywt, gev, loose, DIV
    integer :: eu(2)

    error = 1

    GAM0 = 0.0d0
    GAM1 = 0.0d0
    PSI = 0.0d0
    PPI = 0.0d0
    C = 0.0d0

    DD2 = 0.0d0

    self%QQ = 0.0d0
    self%ZZ = 0.0d0
    self%HH = 0.0d0

    {sims_mat}

    self%DD = DD2(:,1)
    call do_gensys(self%TT, CC, self%RR, fmat, fwt, ywt, gev, eu, loose, GAM0, GAM1, C, PSI, PPI, DIV)



    info = eu(1)*eu(2)

    if (info==1) error = 0

  end subroutine system_matrices


end module model_t
"""


def smc(model, t0=0):

    import sympy
    from sympy.printing import fcode
    cmodel = model.compile_model()
    template = fortran_model #open('fortran_model.f90').read()
    write_prior_file(cmodel.prior,'.')

    system_matrices = model.python_sims_matrices(matrix_format='symbolic')
    npara = len(model.parameters)
    para = sympy.IndexedBase('para',shape=(npara+1,))

    fortran_subs = dict(zip([sympy.symbols('garbage') ]+model.parameters, para))
    fortran_subs[0] = 0.0
    fortran_subs[1] = 1.0
    fortran_subs[100] = 100.0
    fortran_subs[2] = 2.0
    fortran_subs[400] = 400.0
    fortran_subs[4] = 4.0

    from .symbols import Parameter
    context_tuple = ([(p, Parameter(p)) for p in model.parameters]
                     + [(p.name, p) for p in model['other_para']])

    context = dict(context_tuple)
    context['exp'] = sympy.exp
    context['log'] = sympy.log

    to_replace = {}
    for p in model['other_para']:
        to_replace[p] = eval(str(model['para_func'][p.name]), context)


    to_replace = list(to_replace.items())
    print(to_replace)
    from itertools import combinations, permutations

    edges = [(i,j) for i,j in permutations(to_replace,2) if type(i[1]) not in [float,int] and i[1].has(j[0])]

    from sympy import default_sort_key, topological_sort
    para_func = topological_sort([to_replace, edges], default_sort_key)

    to_write = ['GAM0','GAM1','PSI','PPI','self%QQ','DD2','self%ZZ','self%HH']
    fmats = [fcode((mat.subs(para_func)).subs(fortran_subs), assign_to=n, source_format='free', standard=95, contract=False)
	     for mat, n in zip(system_matrices, to_write)]
    sims_mat = '\n\n'.join(fmats)
    template = template.format(model=model, yy=cmodel.yy, p0='',t0=t0, sims_mat=sims_mat)

    return template


def translate(model, output_dir='.', language='fortran'):
    """

    Inputs
    ------
    model - DSGEModel 
    output_dir - directory to write the model to 
    language - language to write

    Returns
    -------
    None 
    """
    if not os.path.isabs(output_dir): 
        output_dir = os.path.join(os.getcwd(), output_dir)

    try: 
        os.mkdir(output_dir)
        print("Created directory: ")
        print("\t", output_dir)
    except:
        print("Directory already exists.")

    
    if language=='fortran':
        translate_fortran(model, output_dir)
    elif language=='matlab':
        translate_matlab(model, output_dir)
    elif language=='dynare':
        raise NotImplementedError('Dynare not yet implemented.')
    elif language=='julia':
        raise NotImplementedError('Julia not yet implemented.')
    else:
        raise ValueError('Unsupported language.')

        
def write_prior_file(prior, output_dir):

    def return_stats(dist):
        if dist.name=='uniform':
            return pdict[dist.name], dist.kwds['loc'], dist.kwds['loc']+dist.kwds['scale'], 0, 0
        elif dist.name=='inv_gamma':
            return pdict[dist.name], dist.a, dist.b, 0, 0
        else:
            return pdict[dist.name], dist.stats()[0], np.sqrt(dist.stats()[1]), 0, 0

    with open(os.path.join(output_dir, 'prior.txt'), mode='w') as prior_file:
        plist = [', '.join(map(str, return_stats(pr))) 
                 for pr in prior.priors]
        prior_file.write('\n'.join(plist))
            

def make_fortran_model(model, **kwargs):
    t0 = kwargs.pop('t0',0)
    from fortress import make_smc

    model_file = smc(model, t0=t0)
    modelc = model.compile_model()

    r = make_smc(model_file, other_files={'data.txt': modelc.yy,
                                          'prior.txt': 'prior.txt'}, **kwargs)

    output_dir = kwargs.pop('output_directory','_fortress_tmp')
    write_prior_file(modelc.prior, output_dir)
    return r
    

def write_trans_file(prior, output_dir):

    def return_trans(dist):
        if dist.name=='uniform':
            return 1, dist.kwds['loc'], dist.kwds['loc']+dist.kwds['scale'], 1
        elif dist.name=='gamma' or dist.name=='inv_gamma':
            return 2, 0, 999, 1
        elif dist.name=='norm':
            return 0, -999, 999, 1
        elif dist.name=='beta':
            return 1, 0, 0.9999, 1
        else:
            raise ValueError("Unable to determine parameter value.")

    with open(os.path.join(output_dir, 'trans.txt'), mode='w') as trans_file:
        plist = [', '.join(map(str, return_trans(pr))) 
                 for pr in prior.priors]
        trans_file.write('\n'.join(plist))


fortran_template = """
!------------------------------------------------------------
! Automatically generated fortran file.
!------------------------------------------------------------
module {name}

  use prior
  use filter
  use gensys


  implicit none

  character(len=*), parameter :: mname = '{name}'

  character(len=*), parameter :: priorfile = '{odir}/prior.txt'
  character(len=*), parameter :: priorfile2 = ''
  character(len=*), parameter :: datafile = '{odir}/yy.txt'
  character(len=*), parameter :: transfile = '{odir}/trans.txt'

  character(len=*), parameter :: initfile = ''
  character(len=*), parameter :: initwt = ''

  character(len=*), parameter :: varfile = ''

  integer, parameter :: neq = {neq}, neta = {neta}, neps = {neps}, nobs = {nobs}, npara = {npara}, ns = {neq}, ny = {ny}
  integer, parameter :: t0 = {t0}

  double precision,  parameter :: REALLY_NEG = -1000000000000.0d0
  ! data
  double precision :: YY(ny, nobs)

  ! prior stuff
  integer :: pshape(npara), pmask(npara)
  double precision :: pmean(npara), pstdd(npara), pfix(npara)

  integer :: pshape2(npara), pmask2(npara)
  double precision :: pmean2(npara), pstdd2(npara), pfix2(npara)

  ! prtr
  double precision :: trspec(4, npara)

contains

  {extra_includes}
  include '/mq/home/m1eph00/code/fortran/base/helper_functions.f90'

  subroutine sysmat(para, TT, RR, QQ, DD, ZZ, HH, info)

    double precision, intent(in) :: para(npara)

    double precision, intent(out) :: TT(neq, neq), RR(neq, neps), QQ(neps, neps), DD(ny), ZZ(ny, neq), HH(ny,ny)
    integer, intent(out) :: info

    double precision :: GAM0(neq, neq), GAM1(neq, neq), C(neq), PSI(neq, neps), PPI(neq, neta), CC(neq)

    double precision :: {para_list}
    double precision :: {helper_list}

    ! gensys
    double precision :: fmat, fwt, ywt, gev, loose, DIV
    integer :: eu(2)

    {para}

    {deff}

    GAM0 = 0.0d0
    GAM1 = 0.0d0
    PSI = 0.0d0
    PPI = 0.0d0
    C = 0.0d0

    {sims_mat}

    call do_gensys(TT, CC, RR, fmat, fwt, ywt, gev, eu, loose, GAM0, GAM1, C, PSI, PPI, DIV)

    info = eu(1)*eu(2)

    QQ = 0.0d0
    ZZ = 0.0d0
    DD = 0.0d0
    HH = 0.0d0

    {ss_mat}

    return
  end subroutine sysmat

  {pmsv}

end module {name}
"""


def write_model_file(model, output_dir, language='fortran', nobs=None):
    
    system_matrices = model.python_sims_matrices(matrix_format='symbolic')
    GAM0, GAM1, PSI, PPI, QQ, DD, ZZ, HH = system_matrices

    from FCodePrinter import fcode_double as wf
    from FCodePrinter import fcode
    from sympy import MatrixSymbol as MS
    sims_mat = '\n\n'.join([wf(eval(mat), MS(mat, *eval(mat).shape), source_format='free', standard=95)
                            for mat in ['GAM0', 'GAM1', 'PSI', 'PPI']])

    ss_mat = '\n\n'.join([wf(eval(mat), MS(mat, *eval(mat).shape), source_format='free', standard=95)
                          for mat in ['QQ', 'DD', 'ZZ', 'HH']])

    para = '\n'.join(['{:} = para({:})'.format(pa, i+1) 
                      for i, pa in enumerate(model.parameters)])

    helper = '\n'.join([wf(model['para_func'][v.name], v.name) for v in model['other_para']])

    

    import re
    cal_list = [str(model['calibration'][str(x)]) for x in model.parameters]
    cal_list = [re.sub(r'([^a-zA-Z][0-9\.]+)', r'\1d0',str_p) for str_p in cal_list]
    pmsv = 'function pmsv result(para)\n  real(wp) :: para(npara)\n\n'
    pmsv += "para = (/" + ",".join(cal_list) + "/)\n\n end function pmsv"


    out = fortran_template.format(odir=output_dir, sims_mat=sims_mat, ss_mat=ss_mat, 
                                  para_list=','.join(map(str, model.parameters)), 
                                  helper_list=','.join(map(str, model['other_para'])), 
                                  para=para+helper, neq=model.neq_fort, neps=model.neps, 
                                  neta=model.neta, nobs=nobs, ny=len(model['observables']), 
                                  npara=len(model.parameters), t0=0, extra_includes='', 
                                  deff='', pmsv=pmsv, **model)

    print(out)
    
def translate_fortran(model, output_dir):

    for driver, driver_file in fortran_files.iteritems():
        with open(os.path.join(template_path, driver_file)) as template:
            if 'Makefile' in driver_file:
                output_path = os.path.join(output_dir, 'Makefile')
            else:
                output_path = os.path.join(output_dir, os.path.basename(driver_file))
            output_file = open(output_path, mode='w')
            output_file.write(template.read().format(model=model['name']))
            output_file.close()

    compiled_model = model.compile_model()
    output_dir = os.path.join(output_dir, 'model')
    try:
        os.mkdir(output_dir)
    except:
        pass

    write_prior_file(compiled_model.prior, output_dir)
    write_trans_file(compiled_model.prior, output_dir)
    write_model_file(model, output_dir, language='fortran', nobs=compiled_model.yy.shape[0])

    np.savetxt(os.path.join(output_dir, 'yy.txt'), compiled_model.yy)
    



# pure python implementation of GENSYS by Chris Sims
import warnings
import numpy as np
from scipy.linalg import ordqz, svd

def gensys(G0, G1, PSI, PI, DIV=1 + 1e-8,
           REALSMALL=1e-6,
           return_everything=False):
    """
    Solves a Linear Rational Expectations model via GENSYS.

    Γ₀xₜ = Γ₁xₜ₋₁ + Ψεₜ + Πηₜ

    Returns
    -------

    RC : 2d int array
         [ 1,  1] = existence and uniqueness
         [ 1,  0] = existence, not uniqueness
         [-2, -2] = coincicdent zeros

    Notes
    -----
    The solution method is detailed in ...

    """
    n, pin = G0.shape[0], PI.shape[1]

    with np.errstate(invalid='ignore', divide='ignore'):
        AA, BB, alpha, beta, Q, Z = ordqz(G0, G1, sort='ouc', output='complex')
        zxz = ((np.abs(beta) < REALSMALL) * (np.abs(alpha) < REALSMALL)).any()

        x = alpha / beta
        nunstab = (x * x.conjugate() < 1.0).sum()

    if zxz:
        RC = [-2, -2]
        print("Coincident zeros")
        return

    nstab = n - nunstab

    Q = Q.T.conjugate()
    Qstab, Qunstab = Q[:nstab, :], Q[nstab:, :]

    etawt = Qunstab.dot(PI)
    ueta, deta, veta = svd(etawt, full_matrices=False)

    bigev = deta > REALSMALL
    deta = deta[bigev]
    ueta = ueta[:, bigev]
    veta = veta[bigev, :].conjugate().T

    RC = np.array([0, 0])
    RC[0] = len(bigev) >= nunstab

    if RC[0] == 0:
        warnings.warn(
            f"{nunstab} unstable roots, but only {len(bigev)} "
            " RE errors! No solution.")

    if nunstab == n:
        raise NotImplementedError("case nunstab == n, not implemented")
    else:
        etawt1 = Qstab.dot(PI)
        ueta1, deta1, veta1 = svd(etawt1, full_matrices=False)
        bigev = deta1 > REALSMALL
        deta1 = deta1[bigev]
        ueta1 = ueta1[:, bigev]
        veta1 = veta1[bigev, :].conjugate().T

    if veta1.size == 0:
        unique = 1
    else:
        loose = veta1 - veta.dot(veta.conjugate().T).dot(veta1)
        ul, dl, vl = svd(loose)
        unique = (dl < REALSMALL).all()

        # existence for general epsilon[t]
        AA22 = AA[-nunstab:, :][:, -nunstab:]
        BB22 = BB[-nunstab:, :][:, -nunstab:]
        M = np.linalg.inv(BB22).dot(AA22)

    if unique:
        RC[1] = 1
    else:
        pass
        # print("Indeterminancy")

    deta = np.diag(1.0 / deta)
    deta1 = np.diag(deta1)

    etawt_inverseT = ueta.dot((veta.dot(deta)).conjugate().T)
    etatw1_T = veta1.dot(deta1).dot(ueta1.conjugate().T)
    tmat = np.c_[np.eye(nstab), -(etawt_inverseT.dot(etatw1_T)).conjugate().T]

    G0 = np.r_[tmat.dot(AA), np.c_[np.zeros(
        (nunstab, nstab)), np.eye(nunstab)]]
    G1 = np.r_[tmat.dot(BB), np.zeros((nunstab, n))]

    G0i = np.linalg.inv(G0)
    G1 = G0i.dot(G1)

    impact = G0i.dot(np.r_[tmat.dot(Q).dot(
        PSI), np.zeros((nunstab, PSI.shape[1]))])

    G1 = np.real(Z.dot(G1).dot(Z.conjugate().T))
    impact = np.real(Z.dot(impact))

    if return_everything:
        GZ = -np.linalg.inv(BB22).dot(Qunstab).dot(PSI)
        GY = Z.dot(G0i[:, -nunstab:])

        return G1, impact, M, GZ, GY, RC

    else:
        return G1, impact, RC



import numpy as np
from numpy.testing import assert_equal

from unittest import TestCase

from dsge.DSGE import DSGE
from dsge.translate import translate

import pkg_resources

class TestTranslate(TestCase):



    def test_simple(self):
        relative_loc = 'examples/sw/sw.yaml'
        model_file = pkg_resources.resource_filename('dsge', relative_loc)
        sw = DSGE.read(model_file)
        #translate(sw, output_dir='/mq/home/m1eph00/tmp/sw_test/')
        pass



from __future__ import division
import numpy as np

from scipy.special import gammaln
import scipy.stats as rv

class InvGamma(object):
    """
    An inverse gamma random variable as detailed in Zellner (1971). 

    A variable sigma follows an inverse gamma (s, nu) if it has the pdf:

    \begin{align}
    p(\sigma|s,\nu) = \frac{2}{\Gamma(\nu/2)} \left(\frac{\nu s^2}{2}\right)^{\nu/2} \frac{1}{\sigma^{\nu+1}} e^{-\nu s^2 / (2\sigma^2)}. 
    \end{align}

    """
    name = 'inv_gamma'

    def __init__(self, s, nu):

        self.a = s
        self.b = nu

    def logpdf(self, x):
        a = self.a
        b = self.b
        if x < 0:
            return -1000000000000

        lpdf = (np.log(2) - gammaln(b/2) + b/2*np.log(b*a**2/2)
                -(b+1)/2*np.log(x**2) - b*a**2/(2*x**2))
        return lpdf
        
    
    def rvs(self):
        rn = rv.norm.rvs(size=(int(self.b), 1))
        return np.sqrt(self.b*self.a**2 / np.sum(rn**2, 0))


class InvGamma1(object):
    """Inverse Gamma 1 distribution
    X ~ IG1(s, nu) if X = sqrt(Y), where Y ~IG2(s, nu) with Y = INV(Z), Z~Gamma(nu/2, 2/s)
    """
    def __init__(self, s, nu):
    
        self.s = s
        self.nu = nu

    def rvs(self):
        return np.sqrt(1.0/rv.gamma.rvs(self.nu/2.0, scale=2.0/self.s))


    def logpdf(self, x):
        pass



import numpy.distutils.core

from setuptools import find_packages


def build_ext(config):

    config_dict = config.todict()
    config_dict.pop("name")
    return config_dict


if __name__ == "__main__":

    from numpy.distutils.misc_util import Configuration

    config_dict = build_ext(Configuration("dsge", parent_package=None, top_path=None))
    numpy.distutils.core.setup(
        name="dsge",
        version="0.1.1",
        platforms="linux",
        packages=find_packages(),
        test_suite="nose.collector",
        tests_require=["nose"],
        package_data={
            "dsge": [
                "examples/ar1/*",
                "examples/DGS/*",
                "examples/edo/*",
                "examples/nkmp/*",
                "examples/schorf_phillips_curve/*",
                "examples/simple-model/*",
                "examples/sw/*",
            ]
        },
        install_requires=["pandas", "scipy" "sympy", "pyyaml", "numba"],
    )



import numpy as np
from numpy.testing import assert_equal

from unittest import TestCase



import dill
dill.settings['recurse'] = True
#dill.detect.trace(True)

import pkg_resources

from dsge.examples import nkmp as dsge1
p0 = dsge1.p0()

nkmp = dsge1.compile_model()


class TestPickle(TestCase):

    def test_abcd(self):
        dill.dumps(nkmp.log_lik)
        self.assertEqual(0,0)


    def test_parallel(self):
        from pathos.multiprocessing import ProcessPool as Pool
        dill.settings['recurse'] = True        
        p = Pool(4)
        l = lambda x: nkmp.log_lik(x)
        r = p.map(nkmp.prior.logpdf, np.array(4*[p0]))
        r = p.map(l, np.array(1000*[p0]))

        



from sympy.printing.fcode import FCodePrinter
from sympy.printing.codeprinter import Assignment


def _print_Assignment(self, expr):
    from sympy.functions.elementary.piecewise import Piecewise
    from sympy.matrices.expressions.matexpr import MatrixSymbol
    from sympy.tensor.indexed import IndexedBase
    lhs = expr.lhs
    rhs = expr.rhs
    # We special case assignments that take multiple lines
    if isinstance(expr.rhs, Piecewise):
        # Here we modify Piecewise so each expression is now
        # an Assignment, and then continue on the print.
        expressions = []
        conditions = []
        for (e, c) in rhs.args:
            expressions.append(Assignment(lhs, e))
            conditions.append(c)
        temp = Piecewise(*zip(expressions, conditions))
        return self._print(temp)
    elif isinstance(lhs, MatrixSymbol):
        # Here we form an Assignment for each element in the array,
        # printing each one.
        lines = []
        for (i, j) in self._traverse_matrix_indices(lhs):
            temp = Assignment(lhs[i, j], rhs[i, j])
            if not(rhs[i, j] == 0):
                code0 = self._print(temp)
                lines.append(code0)

        return "\n".join(lines)
    elif self._settings["contract"] and (lhs.has(IndexedBase) or
            rhs.has(IndexedBase)):
        # Here we check if there is looping to be done, and if so
        # print the required loops.
        return self._doprint_loops(rhs, lhs)
    else:
        lhs_code = self._print(lhs)
        rhs_code = self._print(rhs)
        return self._get_statement("%s = %s" % (lhs_code, rhs_code))

FCodePrinter._print_Assignment = _print_Assignment

from sympy.printing.fcode import fcode

import sympy 





def fcode_double(x, assign_to=None, **settings):
    
    if isinstance(x, (int, float)):
        return fcode(sympy.Float(x), assign_to=assign_to, **settings)
    else:
        return fcode(x.subs([(si, sympy.Float(si)) for si in x.atoms(sympy.Integer)]), assign_to=assign_to, **settings)


import numpy as np


class Prior(object):
    def __init__(self, individual_prior):
        self.priors = individual_prior
        if self.priors is None:
            self.npara = 0
        else:
            self.npara = len(individual_prior)

    def logpdf(self, para):
        if self.priors is None:
            return None

        ind_density = [x.logpdf(y) for x, y in zip(self.priors, para)]
        ldens = np.sum(ind_density)
        if ldens == -np.inf:
            ldens = -100000000000.0
        return ldens

    def rvs(self, size=None):
        if self.priors is None:
            return None
        if size == None:
            return np.array([x.rvs() for x in self.priors])
        else:
            return np.array([[x.rvs() for x in self.priors] for _ in range(size)])



"""
A module for Linear Gaussian State Space models.


Classes
-------
StateSpaceModel
LinearDSGEModel
"""
import numpy as np
import pandas as p

from scipy.linalg import solve_discrete_lyapunov

from .gensys import gensys
from .filters import chand_recursion, kalman_filter, filter_and_smooth

filt_choices = {"chand_recursion": chand_recursion, "kalman_filter": kalman_filter}


class StateSpaceModel(object):
    r"""
    Object for holding state space model

    .. math::

    s_t &=& T(\theta) s_{t-1} + R(\theta) \epsilon_t,
    \quad \epsilon_t \sim N(0,Q(\theta)) \\

    y_t &=& D(\theta) + Z(\theta) s_{t} + \eta_t,
    \quad \eta_t  \sim N(0, H(\theta))

    Attributes
    ----------
    yy : array_like or pandas.DataFrame
        Dataset containing the observables of the model.

    t0 : int, optional
        Number of initial observations to condition on for
        likelihood evaluation. The default is 0.

    shock_names : list or None, optional
        Names of the shocks.

    state_names : list or None, optional
        Names of the states.

    obs_names : list or None, optional
        Names of observables.

    Methods
    -------
    TT(para), RR(para), QQ(para)
        Define the state transition matrices as function of a parameter vector.
    DD(para), ZZ(para), HH(para)
        Define the observable transition matrices as function of a parameter vector.
    log_lik(para)
        Computes the likelihood of the model at parameter value para.
    impulse_response(para, h=20)
        Computes the impulse response function at parameter value para.
    pred(para, h=20, shocks=True, append=False)
        Simulates a draw from the predictive distribution at parameter
        value para.
    kf_everything(para)
        Generates the filtered and smoothed posterior means of the state vector.
    """

    fast_filter = "chand_recursion"

    def __init__(
        self,
        yy,
        CC,
        TT,
        RR,
        QQ,
        DD,
        ZZ,
        HH,
        t0=0,
        shock_names=None,
        state_names=None,
        obs_names=None,
    ):

        if len(yy.shape) < 2:
            yy = np.swapaxes(np.atleast_2d(yy), 0, 1)

        self.yy = yy

        self.CC = CC
        self.TT = TT
        self.RR = RR
        self.QQ = QQ
        self.DD = DD
        self.ZZ = ZZ
        self.HH = HH

        self.t0 = t0

        self.shock_names = shock_names
        self.state_names = state_names
        self.obs_names = obs_names

    def log_lik(self, para, *args, **kwargs):
        """
        Computes the log likelihood of the model.

        Parameters
        ----------
        para : array-like
            An npara length vector of parameter values that defines the system matrices.
        t0 : int, optional
            Number of initial observations to condition on. 
        y : 2d array-like, optional
            Dataset of observables (T x nobs). The default is the observable set pass during
            class instantiation.
        P0 : 2d array-like or string, optional
            [ns x ns] initial covariance matrix of states, or `unconditional` to use the one
            associated with the invariant distribution.  The default is `unconditional.`


        Returns
        -------
        lik : float
            The log likelihood.


        See Also
        --------
        StateSpaceModel.kf_everything

        """

        t0 = kwargs.pop("t0", self.t0)
        yy = kwargs.pop("y", self.yy)
        P0 = kwargs.pop("P0", "unconditional")

        if np.isnan(yy).any().any():
            default_filter = "kalman_filter"
        else:
            default_filter = "chand_recursion"

        filt = kwargs.pop("filter", default_filter)
        filt_func = filt_choices[filt]

        CC, TT, RR, QQ, DD, ZZ, HH = self.system_matrices(para)
        A0 = kwargs.pop("A0", np.zeros(CC.shape))
        if (np.isnan(TT)).any():
            lik = -1000000000000.0
            return lik

        if P0 == "unconditional":
            P0 = solve_discrete_lyapunov(TT, RR.dot(QQ).dot(RR.T))

        lik = filt_func(
            np.asarray(yy),
            CC,
            TT,
            RR,
            QQ,
            np.asarray(DD, dtype=float),
            np.asarray(ZZ, dtype=float),
            np.asarray(HH, dtype=float),
            np.asarray(A0, dtype=float),
            np.asarray(P0, dtype=float),
            t0=t0,
        )
        return lik

    def kf_everything(self, para, *args, **kwargs):
        """
        Runs the kalman filter and returns objects of interest.

        Parameters
        ----------
        para : array-like
            An npara length vector of parameter values that defines the system matrices.
        t0 : int, optional
            Number of initial observations to condition on. 
        y : 2d array-like, optional
            Dataset of observables (T x nobs). The default is the observable set pass during
            class instantiation.
        P0 : 2d arry-like or string, optional
            [ns x ns] initial covariance matrix of states, or `unconditional` to use the one
            associated with the invariant distribution.  The default is `unconditional.`
        shocks : bool, optional
            Whether to filter and smooth for the structural shocks as well as states. 
            The default is True. 

        Returns
        -------
        results : dict of p.DataFrames with
             `log_lik` -- the sequence of log likelihoods
             `filtered_means' -- the filtered means of the states 
             `filtered_std' -- the filtered stds of the states
             `forecast_means' -- the forecasted means of the states 
             `forecast_std' -- the forecasted stds of the states
             `smoothed_means' -- the smoothed means of the model
             `smoothed_stds' -- the smoothed stds of the model 

        Notes
        -----
        Can be used with missing (NaN) observations.
        """

        t0 = kwargs.pop("t0", self.t0)
        yy = kwargs.pop("y", self.yy)
        P0 = kwargs.pop("P0", "unconditional")
        get_shocks = kwargs.pop("shocks", True)
        yy = p.DataFrame(yy)

        CC, TT, RR, QQ, DD, ZZ, HH = self.system_matrices(para, *args, **kwargs)
        if get_shocks:
            from scipy.linalg import block_diag
            neps = RR.shape[1]
            nobs = ZZ.shape[0]
            TT = block_diag(TT, np.zeros((neps, neps)))
            RR = np.vstack([RR, np.eye(neps)])
            CC = np.zeros((TT.shape[0]))
            ZZ = np.hstack([ZZ, np.zeros((nobs, neps))])

        A0 = kwargs.pop("A0", np.zeros(CC.shape))
        if P0 == "unconditional":
            P0 = solve_discrete_lyapunov(TT, RR.dot(QQ).dot(RR.T))

        res = filter_and_smooth(
            np.asarray(yy),
            CC,
            TT,
            RR,
            QQ,
            np.asarray(DD, dtype=float),
            np.asarray(ZZ, dtype=float),
            np.asarray(HH, dtype=float),
            np.asarray(A0, dtype=float),
            np.asarray(P0, dtype=float),
            t0=t0,
        )

        (
            loglh,
            filtered_means,
            filtered_stds,
            filtered_cov,
            forecast_means,
            forecast_stds,
            forecast_cov,
            smoothed_means,
            smoothed_stds,
            smoothed_cov,
        ) = res

        results = {}
        results["log_lik"] = p.DataFrame(loglh, columns=["log_lik"], index=yy.index)

        for resname, res in [
            ("filtered_means", filtered_means),
            ("filtered_stds", filtered_stds),
            ("forecast_means", forecast_means),
            ("forecast_stds", forecast_stds),
            ("smoothed_means", smoothed_means),
            ("smoothed_stds", smoothed_stds),
        ]:

            if get_shocks:
                names = self.state_names + self.shock_names 
            else:
                names = self.state_names
            resdf = p.DataFrame(res, columns=names, index=yy.index)
            results[resname] = resdf

        return results

    def pred(self, para, h=20, shocks=True, append=False, *args, **kwargs):
        """
        Draws from the predictive distribution $p(Y_{t+1:t+h}|Y_{1:T}, \theta)$.


        Parameters
        ----------
        para : array-like
            An npara length vector of parameter values that defines the system matrices.
        h : int, optional
            The horizon of the distribution.
        y : 2d array-like, optional
            Dataset of observables (T x nobs). The default is the observable set pass during
            class instantiation.
        append : bool, optional
            Return the draw appended to yy (default = FALSE).

        Returns
        -------
        ysim : pandas.DataFrame
            A dataframe containing the draw from the predictive distribution.

        """
        yy = kwargs.pop("y", self.yy)
        res = self.kf_everything(para, y=yy, shocks=False, *args, **kwargs)

        CC, TT, RR, QQ, DD, ZZ, HH = self.system_matrices(para, *args, **kwargs)

        At = res["smoothed_means"].iloc[-1].values
        ysim = np.zeros((h, DD.size))

        index0 = res["smoothed_means"].index[-1] + 1
        index = []
        for i in range(h):
            e = np.random.multivariate_normal(np.zeros((QQ.shape[0])), QQ)
            At = CC + TT.dot(At) + shocks * RR.dot(e)
            h = np.random.multivariate_normal(np.zeros((HH.shape[0])), HH)
            At = np.asarray(At).squeeze()
            ysim[i, :] = DD.T + ZZ.dot(At) + shocks * np.atleast_2d(h)
            index.append(index0 + i)

        ysim = p.DataFrame(ysim, columns=self.obs_names, index=index)

        if append:
            ysim = self.yy.append(ysim)
        return ysim

    def system_matrices(self, para, *args, **kwargs):
        """
        Returns the system matrices of the state space model.

        Parameters
        ----------
        para : array-like
            An npara length vector of parameter values that defines the system matrices.


        Returns
        -------
        CC : np.array (ns)
        TT : np.array (ns x ns)
        RR : np.array (ns x neps)
        QQ : np.array (neps x neps)
        DD : np.array (nobs)
        ZZ : np.array (ny x ns)
        HH : np.array (ny x ny)


        Notes
        -----

        """
        CC = np.atleast_1d(self.CC(para, *args, **kwargs))
        TT = np.atleast_2d(self.TT(para, *args, **kwargs))
        RR = np.atleast_2d(self.RR(para, *args, **kwargs))
        QQ = np.atleast_2d(self.QQ(para, *args, **kwargs))

        DD = np.atleast_1d(self.DD(para, *args, **kwargs))
        ZZ = np.atleast_2d(self.ZZ(para, *args, **kwargs))
        HH = np.atleast_1d(self.HH(para, *args, **kwargs))

        return CC, TT, RR, QQ, DD, ZZ, HH

    def abcd_representation(self, para, *args, **kwargs):
        """
        Returns ABCD representation of state space system.

        Parameters
        ----------
        para : array-like
            An npara length vector of parameter values that defines the system matrices.
        

        Returns 
        -------
        A, B, C, D


        Notes
        -----
        

        """

        CC, TT, RR, QQ, DD, ZZ, HH = self.system_matrices(para, *args, **kwargs)

        if not (np.all(HH == 0)):
            fdkjsl

        A = TT
        B = RR

        C = ZZ.dot(TT)  # ZZ @ TT
        D = ZZ.dot(RR)

        return A, B, C, D

    def impulse_response(self, para, h=20, *args, **kwargs):
        """
        Computes impulse response functions of model.

        Parameters
        ----------
        para : array-like
            An npara length vector of parameter values that defines the system matrices.

        h : int, optional
           The maximum horizon length for the impulse responses.


        Returns
        -------
        irf : pandas.Panel (nshocks x h x nvariables)


        Notes
        -----
        These are of the model state variables impulse responses to
        1 standard deviation shocks.  The method does not return IRFs in terms
        of the model observables.
        """
        CC, TT, RR, QQ, DD, ZZ, HH = self.system_matrices(para, *args, **kwargs)

        if self.shock_names == None:
            self.shock_names = ["shock_" + str(i) for i in range(QQ.shape[0])]
            self.state_names = ["state_" + str(i) for i in range(TT.shape[0])]
            self.obs_names = ["obs_" + str(i) for i in range(ZZ.shape[0])]

        neps = QQ.shape[0]

        irfs = {}
        for i in range(neps):

            At = np.zeros((TT.shape[0], h + 1))
            QQz = np.zeros_like(QQ)
            QQz[i, i] = QQ[i, i]
            cQQz = np.sqrt(QQz)

            # cQQz = np.linalg.cholesky(QQz)

            At[:, 0] = (RR.dot(cQQz)[:, i]).squeeze()

            for j in range(h):
                At[:, j + 1] = TT.dot(At[:, j])

            irfs[self.shock_names[i]] = p.DataFrame(At.T, columns=self.state_names)

        return irfs

    def simulate(self, para, nsim=200, *args, **kwargs):
        """
        Simulates the observables of the model.

        Parameters
        ----------
        para : array-like
            An npara length vector of parameter values that defines the system matrices.

        nsim : int, optional
            The length of the simulation. The default value is 200.

        Returns
        -------
        ysim : np.array (nsim x nobs)


        Notes
        -----
        The simulation is initialized from the steady-state mean and subsequently
        a simulation of length 2*nsim is created, with the final nsim observations
        return.
        """

        CC, TT, RR, QQ, DD, ZZ, HH = self.system_matrices(para, *args, **kwargs)
        ysim = np.zeros((nsim * 2, DD.size))
        At = np.zeros((TT.shape[0],))

        for i in range(nsim * 2):
            e = np.random.multivariate_normal(np.zeros((QQ.shape[0])), QQ)
            At = CC + TT.dot(At) + RR.dot(e)

            h = np.random.multivariate_normal(np.zeros((HH.shape[0])), HH)
            At = np.asarray(At).squeeze()
            ysim[i, :] = DD.T + ZZ.dot(At) + np.atleast_2d(h)

        return ysim[nsim:, :]

    def forecast(self, para, h=20, shocks=True, *args, **kwargs):
        t0 = kwargs.pop("t0", self.t0)
        yy = kwargs.pop("y", self.yy)
        P0 = kwargs.pop("P0", "unconditional")

        CC, TT, RR, QQ, DD, ZZ, HH = self.system_matrices(para, *args, **kwargs)

        if P0 == "unconditional":
            P0 = solve_discrete_lyapunov(TT, RR.dot(QQ).dot(RR.T))

        data = np.asarray(yy)
        nobs, ny = yy.shape
        ns = TT.shape[0]

        At = np.zeros(shape=(nobs, ns))

        RQR = np.dot(np.dot(RR, QQ), RR.T)

        Pt = P0

        ZZ = np.atleast_2d(ZZ)
        HH = np.atleast_2d(HH)
        loglh = np.zeros((nobs))
        AA = At[0, :]

        for i in np.arange(0, nobs):

            At[i, :] = AA

            yhat = np.dot(ZZ, AA) + DD.flatten()
            nut = data[i, :] - yhat
            nut = np.atleast_2d(nut)
            ind = (np.isnan(data[i, :]) == False).nonzero()[0]
            nyact = ind.size

            if nyact > 0:
                Ft = np.dot(np.dot(ZZ[ind, :], Pt), ZZ[ind, :].T) + HH[ind, :][:, ind]
                Ft = 0.5 * (Ft + Ft.T)

                dFt = np.log(np.linalg.det(Ft))

                iFtnut = sp.linalg.solve(Ft, nut[:, ind].T, sym_pos=True)

                if i + 1 > t0:
                    loglh[i] = (
                        -0.5 * nyact * np.log(2 * np.pi)
                        - 0.5 * dFt
                        - 0.5 * np.dot(nut[:, ind], iFtnut)
                    )

                TTPt = np.dot(TT, Pt)
                Kt = np.dot(TTPt, ZZ[ind, :].T)
            else:
                TTPt = np.dot(TT, Pt)

                Kt = np.zeros((ns, ny))
                Ft = np.eye(ny)
                iFtnut = np.eye(ny)

            AA = CC + np.dot(TT, AA) + np.dot(Kt, iFtnut).squeeze()
            AA = np.asarray(AA).squeeze()

            Pt = (
                np.dot(TTPt, TT.T)
                - np.dot(Kt, sp.linalg.solve(Ft, Kt.T, sym_pos=True))
                + RQR
            )

        if isinstance(yy, p.DataFrame):
            loglh = p.DataFrame(loglh, columns=["Log Lik."])
            loglh.index = yy.index

        results = {}
        results["filtered_states"] = p.DataFrame(
            At, columns=self.state_names, index=yy.index
        )
        results["one_step_forecast"] = []
        results["log_lik"] = loglh

        return results

    def historical_decomposition(self, para, *args, **kwargs):
        pass


class LinearDSGEModel(StateSpaceModel):
    def __init__(
        self,
        yy,
        GAM0,
        GAM1,
        PSI,
        PPI,
        QQ,
        DD,
        ZZ,
        HH,
        t0=0,
        shock_names=None,
        state_names=None,
        obs_names=None,
        prior=None,
    ):

        if len(yy.shape) < 2:
            yy = np.swapaxes(np.atleast_2d(yy), 0, 1)

        self.yy = yy

        self.GAM0 = GAM0
        self.GAM1 = GAM1
        self.PSI = PSI
        self.PPI = PPI
        self.QQ = QQ
        self.DD = DD
        self.ZZ = ZZ
        self.HH = HH

        self.t0 = t0

        self.shock_names = shock_names
        self.state_names = state_names
        self.obs_names = obs_names

        self.prior = prior

    def solve_LRE(self, para, *args, **kwargs):

        G0 = self.GAM0(para, *args, **kwargs)
        G1 = self.GAM1(para, *args, **kwargs)
        PSI = self.PSI(para, *args, **kwargs)
        PPI = self.PPI(para, *args, **kwargs)

        G0 = np.atleast_2d(G0)
        G1 = np.atleast_2d(G1)
        PSI = np.atleast_2d(PSI)
        PPI = np.atleast_2d(PPI)

        C0 = np.zeros(G0.shape[0])

        nf = PPI.shape[1]

        if nf > 0:
            TT, RR, RC = gensys(G0, G1, PSI, PPI, C0)
            RC = RC[0] * RC[1]
            # TT, CC, RR, fmat, fwt, ywt, gev, RC, loose = gensysw.gensys.call_gensys(G0, G1, C0, PSI, PPI, 1.00000000001)
        else:
            TT = np.linalg.inv(G0).dot(G1)
            RR = np.linalg.inv(G0).dot(PSI)
            RC = 1

        return TT, RR, RC

    def system_matrices(self, para, *args, **kwargs):

        TT, RR, RC = self.solve_LRE(para, *args, **kwargs)
        CC = np.zeros(TT.shape[0])

        QQ = np.atleast_2d(self.QQ(para, *args, **kwargs))
        DD = np.atleast_1d(self.DD(para, *args, **kwargs))
        ZZ = np.atleast_2d(self.ZZ(para, *args, **kwargs))
        HH = np.atleast_1d(self.HH(para, *args, **kwargs))

        if RC != 1:
            TT = np.nan * TT

        return CC, TT, RR, QQ, DD, ZZ, HH

    def log_pr(self, para, *args, **kwargs):
        try:
            return self.prior.logpdf(para)
        except:
            pass
            # raise("no prior specified")

    def log_post(self, para, *args, **kwargs):
        x = self.log_lik(para) + self.log_pr(para)
        if np.isnan(x):
            x = -1000000000.0
        if x < -1000000000.0:
            x = -1000000000.0
        return x
